@@ -42,7 +42,7 @@ static char sccsid[] = "@(#)mkfs.c	8.11 (Berkeley) 5/3/95";
 #endif /* not lint */
 #endif
 #include <sys/cdefs.h>
-__FBSDID("$FreeBSD: src/sbin/newfs/mkfs.c,v 1.92 2006/10/31 21:52:27 pjd Exp $");
+__FBSDID("$FreeBSD: src/sbin/newfs/mkfs.c,v 1.92.8.1 2009/04/15 03:14:26 kensmith Exp $");
 
 #include <err.h>
 #include <grp.h>
